@@ -15,7 +15,7 @@
 
 """This module contains the TestCase class and other helper code, like decorators for test methods."""
 
-### TODO: finish doing the retry stuff for the inner clauses
+# TODO: finish doing the retry stuff for the inner clauses
 
 from __future__ import with_statement
 
@@ -35,7 +35,6 @@ import deprecated_assertions
 from testify.utils import class_logger
 from testify.utils import inspection
 
-# just a useful list to have
 FIXTURE_TYPES = (
     'class_setup',
     'setup',
@@ -44,11 +43,15 @@ FIXTURE_TYPES = (
     'setup_teardown',
     'class_setup_teardown',
 )
+FIXTURES_WHICH_CAN_RETURN_UNEXPECTED_RESULTS = (
+    'class_teardown',
+    'class_setup_teardown',
+)
 
-# in general, inherited fixtures are applied first unless they are of these
-# types. these fixtures are applied (in order of their definitions) starting
-# with those defined on the current class, and and then those defined on
-# inherited classes (following MRO).
+# In general, inherited fixtures are applied first unless they are of these
+# types. These fixtures are applied (in order of their definitions) starting
+# with those defined on the current class, and then those defined on inherited
+# classes (following MRO).
 REVERSED_FIXTURE_TYPES = (
     'teardown',
     'class_teardown',
@@ -72,7 +75,7 @@ class TwistedFailureError(Exception):
 
 class MetaTestCase(type):
     """This base metaclass is used to collect each TestCase's decorated fixture methods at
-    runtime.  It is implemented as a metaclass so we can determine the order in which
+    runtime. It is implemented as a metaclass so we can determine the order in which
     fixture methods are defined.
     """
     __test__ = False
@@ -102,8 +105,7 @@ class MetaTestCase(type):
         else:
             return hash(MetaTestCase._cmp_str(self)) % bucket_count
 
-def discovered_test_cases():
-    return [test_case_class for test_case_class in MetaTestCase._test_accumulator if test_case_class != TestCase]
+
 
 class TestCase(object):
     """The TestCase class defines test methods and fixture methods; it is the meat and potatoes of testing.
@@ -148,8 +150,8 @@ class TestCase(object):
     EVENT_ON_COMPLETE_CLASS_SETUP_METHOD = 4
     EVENT_ON_RUN_CLASS_TEARDOWN_METHOD = 5
     EVENT_ON_COMPLETE_CLASS_TEARDOWN_METHOD = 6
-    EVENT_ON_RUN_FIXTURE_METHOD = 7
-    EVENT_ON_COMPLETE_FIXTURE_METHOD = 8
+    EVENT_ON_RUN_TEST_CASE = 7
+    EVENT_ON_COMPLETE_TEST_CASE = 8
 
     log = class_logger.ClassLogger()
 
@@ -236,7 +238,7 @@ class TestCase(object):
                 )
 
                 # we grabbed this from the class and need to bind it to us
-                instance_method = instancemethod(unbound_method, self)
+                instance_method = instancemethod(unbound_method, self, self.__class__)
                 self._fixture_methods[instance_method._fixture_type].append(instance_method)
 
         # arrange our fixture buckets appropriately
@@ -273,7 +275,9 @@ class TestCase(object):
             member = getattr(self, member_name)
             if not inspect.ismethod(member):
                 continue
-            member_suites = getattr(member, '_suites', set()) | set(getattr(self, '_suites', []))
+
+            member_suites = self.suites(member)
+
             # if there are any exclude suites, exclude methods under them
             if self.__suites_exclude and self.__suites_exclude & member_suites:
                 continue
@@ -289,12 +293,32 @@ class TestCase(object):
                 yield member
 
     def run(self):
-        """Delegator method encapsulating the flow for executing a TestCase instance"""
+        """Delegator method encapsulating the flow for executing a TestCase instance.
+
+        This method tracks its progress in a TestResult with test_method 'run'.
+        This TestResult is used as a signal when running in client/server mode:
+        when the client is done running a TestCase and its fixtures, it sends
+        this TestResult to the server during the EVENT_ON_COMPLETE_TEST_CASE
+        phase.
+
+        This could be handled better. See
+        https://github.com/Yelp/Testify/issues/121.
+        """
+
+        # The TestResult constructor wants an actual method, which it inspects
+        # to determine the method name (and class name, so it must be a method
+        # and not a function!). self.run is as good a method as any.
+        test_case_result = TestResult(self.run)
+        test_case_result.start()
+        self.fire_event(self.EVENT_ON_RUN_TEST_CASE, test_case_result)
 
         self.__run_class_setup_fixtures()
-        self.__enter_context_managers(self.class_setup_teardown_fixtures, self.__run_test_methods)
+        self.__enter_class_context_managers(self.class_setup_teardown_fixtures, self.__run_test_methods)
         self.__run_class_teardown_fixtures()
 
+        test_case_result.end_in_success()
+        self.fire_event(self.EVENT_ON_COMPLETE_TEST_CASE, test_case_result)
+
     def __run_class_setup_fixtures(self):
         """Running the class's class_setup method chain."""
         self.__run_class_fixtures(
@@ -321,25 +345,30 @@ class TestCase(object):
             result = TestResult(fixture_method)
 
             try:
-                for callback in self.__callbacks[callback_on_run_event]:
-                    callback(result.to_dict())
-
                 result.start()
-
+                self.fire_event(callback_on_run_event, result)
                 if self.__execute_block_recording_exceptions(fixture_method, result, is_class_level=True):
                     result.end_in_success()
+                else:
+                    self.failure_count += 1
             except (KeyboardInterrupt, SystemExit):
                 result.end_in_interruption(sys.exc_info())
                 raise
             finally:
-                for callback in self.__callbacks[callback_on_complete_event]:
-                    callback(result.to_dict())
+                self.fire_event(callback_on_complete_event, result)
 
     @classmethod
     def in_suite(cls, method, suite_name):
         """Return a bool denoting whether the given method is in the given suite."""
         return suite_name in getattr(method, '_suites', set())
 
+    def suites(self, method=None):
+        """Returns the suites associated with this test case and, optionally, the given method."""
+        suites = set(getattr(self, '_suites', []))
+        if method is not None:
+            suites |= getattr(method, '_suites', set())
+        return suites
+
     def method_excluded(self, method):
         """Given this TestCase's included/excluded suites, is this test method excluded?
 
@@ -349,6 +378,31 @@ class TestCase(object):
         method_suites = set(getattr(method, '_suites', set()))
         return (self.__suites_exclude & method_suites)
 
+
+    def __enter_class_context_managers(self, fixture_methods, callback):
+        """Transform each fixture_method into a context manager with contextlib.contextmanager, enter them recursively, and call callback"""
+        if fixture_methods:
+            fixture_method = fixture_methods[0]
+            ctm = contextmanager(fixture_method)()
+
+            enter_result = TestResult(fixture_method)
+            enter_result.start()
+            self.fire_event(self.EVENT_ON_RUN_CLASS_SETUP_METHOD, enter_result)
+            if self.__execute_block_recording_exceptions(ctm.__enter__, enter_result, is_class_level=True):
+                enter_result.end_in_success()
+            self.fire_event(self.EVENT_ON_COMPLETE_CLASS_SETUP_METHOD, enter_result)
+
+            self.__enter_context_managers(fixture_methods[1:], callback)
+
+            exit_result = TestResult(fixture_method)
+            exit_result.start()
+            self.fire_event(self.EVENT_ON_RUN_CLASS_TEARDOWN_METHOD, exit_result)
+            if self.__execute_block_recording_exceptions(lambda: ctm.__exit__(None, None, None), exit_result, is_class_level=True):
+                exit_result.end_in_success()
+            self.fire_event(self.EVENT_ON_COMPLETE_CLASS_TEARDOWN_METHOD, exit_result)
+        else:
+            callback()
+
     def __enter_context_managers(self, fixture_methods, callback):
         """Transform each fixture_method into a context manager with contextlib.contextmanager, enter them recursively, and call callback"""
         if fixture_methods:
@@ -365,20 +419,23 @@ class TestCase(object):
         phase, no method-level fixtures or test methods will be run, and we'll eventually
         skip all the way to the class_teardown phase.   If a given test method is marked
         as disabled, neither it nor its fixtures will be run.  If there is an exception
-        during during the setup phase, the test method will not be run and execution
+        during the setup phase, the test method will not be run and execution
         will continue with the teardown phase.
         """
         for test_method in self.runnable_test_methods():
 
             result = TestResult(test_method)
+            # Sometimes, test cases want to take further action based on
+            # results, e.g. further clean-up or reporting if a test method
+            # fails. (Yelp's Selenium test cases do this.)
             test_method.im_self.test_result = result
 
             try:
                 self._method_level = True # Flag that we're currently running method-level stuff (rather than class-level)
 
-                # run "on-run" callbacks. eg/ print out the test method name
-                for callback in self.__callbacks[self.EVENT_ON_RUN_TEST_METHOD]:
-                    callback(result.to_dict())
+                # run "on-run" callbacks. e.g. print out the test method name
+                self.fire_event(self.EVENT_ON_RUN_TEST_METHOD, result)
+
                 result.start()
 
                 if self.__class_level_failure:
@@ -408,10 +465,8 @@ class TestCase(object):
 
                     # finally, run the teardown phase
                     self._stage = self.STAGE_TEARDOWN
-                    def _teardown_block():
-                        for fixture_method in self.teardown_fixtures:
-                            fixture_method()
-                    self.__execute_block_recording_exceptions(_teardown_block, result)
+                    for fixture_method in self.teardown_fixtures:
+                        self.__execute_block_recording_exceptions(fixture_method, result)
 
                 # if nothing's gone wrong, it's not about to start
                 if not result.complete:
@@ -420,9 +475,7 @@ class TestCase(object):
                 result.end_in_interruption(sys.exc_info())
                 raise
             finally:
-                for callback in self.__callbacks[self.EVENT_ON_COMPLETE_TEST_METHOD]:
-                    callback(result.to_dict())
-
+                self.fire_event(self.EVENT_ON_COMPLETE_TEST_METHOD, result)
                 self._method_level = False
 
                 if not result.success:
@@ -440,9 +493,13 @@ class TestCase(object):
         """
         self.__callbacks[event].append(callback)
 
+    def fire_event(self, event, result):
+        for callback in self.__callbacks[event]:
+            callback(result.to_dict())
+
     def __execute_block_recording_exceptions(self, block_fxn, result, is_class_level=False):
-        """Excerpted code for executing a block of code that might except and
-        cause us to update a result object.
+        """Excerpted code for executing a block of code that might raise an
+        exception, requiring us to update a result object.
 
         Return value is a boolean describing whether the block was successfully
         executed without exceptions.
@@ -497,11 +554,14 @@ class TestCase(object):
 class TestifiedUnitTest(TestCase, unittest.TestCase):
 
     @classmethod
-    def from_unittest_case(cls, unittest_class):
+    def from_unittest_case(cls, unittest_class, module_suites=None):
         """"Constructs a new testify.TestCase from a unittest.TestCase class.
 
         This operates recursively on the TestCase's class hierarchy by
         converting each parent unittest.TestCase into a TestifiedTestCase.
+
+        If 'suites' are provided, they are treated as module-level suites to be
+        applied in addition to class- and test-level suites.
         """
 
         # our base case: once we get to the parent TestCase, replace it with our
@@ -526,12 +586,16 @@ class TestifiedUnitTest(TestCase, unittest.TestCase):
         # use an __init__ smart enough to figure out our inheritance
         unittest_dict['__init__'] = cls.__init__
 
+        # add module-level suites in addition to any suites already on the class
+        class_suites = set(getattr(unittest_class, '_suites', []))
+        unittest_dict['_suites'] = class_suites | set(module_suites or [])
+
         # traverse our class hierarchy and 'testify' parent unittest.TestCases
         bases = []
 
         for base_class in unittest_class.__bases__:
             if issubclass(base_class, unittest.TestCase):
-                base_class = cls.from_unittest_case(base_class)
+                base_class = cls.from_unittest_case(base_class, module_suites=module_suites)
             bases.append(base_class)
 
         # include our original unittest class so existing super() calls still
@@ -559,7 +623,7 @@ def suite(*args, **kwargs):
         if not hasattr(function, '_suites'):
             function._suites = set()
         if args and (conditions is None or bool(conditions) is True):
-            function._suites.update(set(args))
+            function._suites = set(function._suites) | set(args)
             if reason:
                 if not hasattr(function, '_suite_reasons'):
                     function._suite_reasons = []
@@ -626,6 +690,7 @@ class_teardown = __fixture_decorator_factory('class_teardown')
 setup_teardown = __fixture_decorator_factory('setup_teardown')
 class_setup_teardown = __fixture_decorator_factory('class_setup_teardown')
 
+
 class let(object):
     """Decorator that creates a lazy-evaluated helper property. The value is
     cached across multiple calls in the same test, but not across multiple
@@ -658,3 +723,4 @@ class let(object):
     def _reset_value(self):
         self._result = self._unsaved
 
+# vim: set ts=4 sts=4 sw=4 et:
